/**
 * data-structure-typed
 *
 * @author Tyler Zeng
 * @copyright Copyright (c) 2022 Tyler Zeng <zrwusa@gmail.com>
 * @license MIT License
 */
import type { BSTNodeKeyOrNode, BTNKey, BTNodeExemplar, TreeMultimapNodeNested, TreeMultimapOptions } from '../../types';
import { BiTreeDeleteResult, BTNCallback, IterationType, TreeMultimapNested } from '../../types';
import { IBinaryTree } from '../../interfaces';
import { AVLTree, AVLTreeNode } from './avl-tree';
export declare class TreeMultimapNode<V = any, N extends TreeMultimapNode<V, N> = TreeMultimapNodeNested<V>> extends AVLTreeNode<V, N> {
    count: number;
    /**
     * The constructor function initializes a BinaryTreeNode object with a key, value, and count.
     * @param {BTNKey} key - The `key` parameter is of type `BTNKey` and represents the unique identifier
     * of the binary tree node.
     * @param {V} [value] - The `value` parameter is an optional parameter of type `V`. It represents the value of the binary
     * tree node. If no value is provided, it will be `undefined`.
     * @param {number} [count=1] - The `count` parameter is a number that represents the number of times a particular value
     * occurs in a binary tree node. It has a default value of 1, which means that if no value is provided for the `count`
     * parameter when creating a new instance of the `BinaryTreeNode` class.
     */
    constructor(key: BTNKey, value?: V, count?: number);
}
/**
 * The only distinction between a TreeMultimap and a AVLTree lies in the ability of the former to store duplicate nodes through the utilization of counters.
 */
export declare class TreeMultimap<V = any, N extends TreeMultimapNode<V, N> = TreeMultimapNode<V, TreeMultimapNodeNested<V>>, TREE extends TreeMultimap<V, N, TREE> = TreeMultimap<V, N, TreeMultimapNested<V, N>>> extends AVLTree<V, N, TREE> implements IBinaryTree<V, N, TREE> {
    constructor(elements?: Iterable<BTNodeExemplar<V, N>>, options?: Partial<TreeMultimapOptions>);
    private _count;
    get count(): number;
    /**
     * The function creates a new BSTNode with the given key, value, and count.
     * @param {BTNKey} key - The key parameter is the unique identifier for the binary tree node. It is used to
     * distinguish one node from another in the tree.
     * @param {N} value - The `value` parameter represents the value that will be stored in the binary search tree node.
     * @param {number} [count] - The "count" parameter is an optional parameter of type number. It represents the number of
     * occurrences of the value in the binary search tree node. If not provided, the count will default to 1.
     * @returns A new instance of the BSTNode class with the specified key, value, and count (if provided).
     */
    createNode(key: BTNKey, value?: V, count?: number): N;
    createTree(options?: TreeMultimapOptions): TREE;
    /**
     * The function checks if an exemplar is an instance of the TreeMultimapNode class.
     * @param exemplar - The `exemplar` parameter is of type `BTNodeExemplar<V, N>`.
     * @returns a boolean value indicating whether the exemplar is an instance of the TreeMultimapNode
     * class.
     */
    isNode(exemplar: BTNodeExemplar<V, N>): exemplar is N;
    /**
     * The function `exemplarToNode` converts an exemplar object into a node object.
     * @param exemplar - The `exemplar` parameter is of type `BTNodeExemplar<V, N>`, where `V` represents
     * the value type and `N` represents the node type.
     * @param [count=1] - The `count` parameter is an optional parameter that specifies the number of
     * times the node should be created. If not provided, it defaults to 1.
     * @returns a value of type `N` (the generic type parameter) or `undefined`.
     */
    exemplarToNode(exemplar: BTNodeExemplar<V, N>, count?: number): N | undefined;
    /**
     * Time Complexity: O(log n) - logarithmic time, where "n" is the number of nodes in the tree. The add method of the superclass (AVLTree) has logarithmic time complexity.
     * Space Complexity: O(1) - constant space, as it doesn't use additional data structures that scale with input size.
     */
    /**
     * Time Complexity: O(log n) - logarithmic time, where "n" is the number of nodes in the tree. The add method of the superclass (AVLTree) has logarithmic time complexity.
     * Space Complexity: O(1) - constant space, as it doesn't use additional data structures that scale with input size.
     *
     * The `add` function overrides the base class `add` function to add a new node to the tree multimap
     * and update the count.
     * @param keyOrNodeOrEntry - The `keyOrNodeOrEntry` parameter can be one of the following:
     * @param [count=1] - The `count` parameter is an optional parameter that specifies the number of
     * times the key or node or entry should be added to the multimap. If not provided, the default value
     * is 1.
     * @returns either a node (`N`) or `undefined`.
     */
    add(keyOrNodeOrEntry: BTNodeExemplar<V, N>, count?: number): N | undefined;
    /**
     * Time Complexity: O(k log n) - logarithmic time, where "n" is the number of nodes in the tree. The add method of the superclass (AVLTree) has logarithmic time complexity.
     * Space Complexity: O(1) - constant space, as it doesn't use additional data structures that scale with input size.
     */
    /**
     * Time Complexity: O(k log n) - logarithmic time, where "n" is the number of nodes in the tree. The add method of the superclass (AVLTree) has logarithmic time complexity.
     * Space Complexity: O(1) - constant space, as it doesn't use additional data structures that scale with input size.
     *
     * The function overrides the addMany method to add multiple keys, nodes, or entries to a data
     * structure.
     * @param keysOrNodesOrEntries - The parameter `keysOrNodesOrEntries` is an iterable that can contain
     * either keys, nodes, or entries.
     * @returns The method is returning an array of type `N | undefined`.
     */
    addMany(keysOrNodesOrEntries: Iterable<BTNodeExemplar<V, N>>): (N | undefined)[];
    /**
     * Time Complexity: O(1) - constant time, as it performs basic pointer assignments.
     * Space Complexity: O(1) - constant space, as it only uses a constant amount of memory.
     */
    /**
     * Time Complexity: O(n log n) - logarithmic time for each insertion, where "n" is the number of nodes in the tree. This is because the method calls the add method for each node.
     * Space Complexity: O(n) - linear space, as it creates an array to store the sorted nodes.
     *
     * The `perfectlyBalance` function takes a sorted array of nodes and builds a balanced binary search
     * tree using either a recursive or iterative approach.
     * @param iterationType - The `iterationType` parameter is an optional parameter that specifies the
     * type of iteration to use when building the balanced binary search tree. It can have two possible
     * values:
     * @returns a boolean value.
     */
    perfectlyBalance(iterationType?: IterationType): boolean;
    /**
     * Time Complexity: O(k log n) - logarithmic time for each insertion, where "n" is the number of nodes in the tree, and "k" is the number of keys to be inserted. This is because the method iterates through the keys and calls the add method for each.
     * Space Complexity: O(1) - constant space, as it doesn't use additional data structures that scale with input size.
     */
    /**
     * Time Complexity: O(log n) - logarithmic time, where "n" is the number of nodes in the tree. The delete method of the superclass (AVLTree) has logarithmic time complexity.
     * Space Complexity: O(1) - constant space, as it doesn't use additional data structures that scale with input size.
     *
     * The `delete` function in TypeScript is used to remove a node from a binary tree, taking into
     * account the count of the node and balancing the tree if necessary.
     * @param identifier - The identifier is the value or key that is used to identify the node that
     * needs to be deleted from the binary tree. It can be of any type that is returned by the callback
     * function.
     * @param {C} callback - The `callback` parameter is a function that is used to determine if a node
     * should be deleted. It is optional and defaults to a default callback function. The `callback`
     * function takes one parameter, which is the identifier of the node, and returns a value that is
     * used to identify the node to
     * @param [ignoreCount=false] - A boolean flag indicating whether to ignore the count of the node
     * being deleted. If set to true, the count of the node will not be considered and the node will be
     * deleted regardless of its count. If set to false (default), the count of the node will be
     * decremented by 1 and
     * @returns an array of `BiTreeDeleteResult<N>`.
     */
    delete<C extends BTNCallback<N>>(identifier: ReturnType<C>, callback?: C, ignoreCount?: boolean): BiTreeDeleteResult<N>[];
    /**
     * Time Complexity: O(n log n) - logarithmic time for each insertion, where "n" is the number of nodes in the tree. This is because the method calls the add method for each node.
     * Space Complexity: O(n) - linear space, as it creates an array to store the sorted nodes.
     */
    /**
     * The clear() function clears the contents of a data structure and sets the count to zero.
     */
    clear(): void;
    /**
     * Time Complexity: O(1) - constant time, as it performs basic pointer assignments.
     * Space Complexity: O(1) - constant space, as it only uses a constant amount of memory.
     *
     * The function adds a new node to a binary tree, either as the left child or the right child of a
     * given parent node.
     * @param {N | undefined} newNode - The `newNode` parameter represents the node that needs to be
     * added to the binary tree. It can be of type `N` (which represents a node in the binary tree) or
     * `undefined` if there is no node to add.
     * @param {BTNKey | N | undefined} parent - The `parent` parameter represents the parent node to
     * which the new node will be added as a child. It can be either a node object (`N`) or a key value
     * (`BTNKey`).
     * @returns The method `_addTo` returns either the `parent.left` or `parent.right` node that was
     * added, or `undefined` if no node was added.
     */
    protected _addTo(newNode: N | undefined, parent: BSTNodeKeyOrNode<N>): N | undefined;
    /**
     * The `_swapProperties` function swaps the key, value, count, and height properties between two nodes.
     * @param {BTNKey | N | undefined} srcNode - The `srcNode` parameter represents the source node from
     * which the values will be swapped. It can be of type `BTNKey`, `N`, or `undefined`.
     * @param {BTNKey | N | undefined} destNode - The `destNode` parameter represents the destination
     * node where the values from the source node will be swapped to.
     * @returns either the `destNode` object if both `srcNode` and `destNode` are defined, or `undefined`
     * if either `srcNode` or `destNode` is undefined.
     */
    protected _swapProperties(srcNode: BSTNodeKeyOrNode<N>, destNode: BSTNodeKeyOrNode<N>): N | undefined;
    protected _replaceNode(oldNode: N, newNode: N): N;
}
