/**
 * data-structure-typed
 *
 * @author Tyler Zeng
 * @copyright Copyright (c) 2022 Tyler Zeng <zrwusa@gmail.com>
 * @license MIT License
 */
import { HashMapLinkedNode, HashMapOptions } from '../../types';
export declare class HashMap<K = any, V = any> {
    protected _noObjMap: Record<string, HashMapLinkedNode<K, V | undefined>>;
    protected _objMap: WeakMap<object, HashMapLinkedNode<K, V | undefined>>;
    protected _head: HashMapLinkedNode<K, V | undefined>;
    protected _tail: HashMapLinkedNode<K, V | undefined>;
    protected readonly _sentinel: HashMapLinkedNode<K, V | undefined>;
    protected _hashFn: (key: K) => string;
    protected _objHashFn: (key: K) => object;
    constructor(elements?: Iterable<[K, V]>, options?: HashMapOptions<K>);
    protected _size: number;
    get size(): number;
    /**
     * Time Complexity: O(1)
     * Space Complexity: O(1)
     *
     * The function returns the key-value pair at the front of a data structure.
     * @returns The front element of the data structure, represented as a tuple with a key (K) and a
     * value (V).
     */
    get first(): [K, V] | undefined;
    /**
     * Time Complexity: O(1)
     * Space Complexity: O(1)
     *
     * The function returns the key-value pair at the end of a data structure.
     * @returns The method is returning an array containing the key-value pair of the tail element in the
     * data structure.
     */
    get last(): [K, V] | undefined;
    /**
     * The `begin()` function in TypeScript iterates over a linked list and yields key-value pairs.
     */
    begin(): Generator<(K | V | undefined)[], void, unknown>;
    /**
     * The function `reverseBegin()` iterates over a linked list in reverse order, yielding each node's
     * key and value.
     */
    reverseBegin(): Generator<(K | V | undefined)[], void, unknown>;
    /**
     * Time Complexity: O(1)
     * Space Complexity: O(1)
     *
     * The `set` function adds a new key-value pair to a data structure, either using an object key or a
     * string key.
     * @param {K} key - The `key` parameter is the key to be set in the data structure. It can be of any
     * type, but typically it is a string or symbol.
     * @param {V} [value] - The `value` parameter is an optional parameter of type `V`. It represents the
     * value associated with the key being set in the data structure.
     * @returns the size of the data structure after the key-value pair has been set.
     */
    set(key: K, value?: V): number;
    /**
     * Time Complexity: O(1)
     * Space Complexity: O(1)
     *
     * The function `get` retrieves the value associated with a given key from a map, either by using the
     * key directly or by using an index stored in the key object.
     * @param {K} key - The `key` parameter is the key used to retrieve a value from the map. It can be
     * of any type, but typically it is a string or symbol.
     * @returns The value associated with the given key is being returned. If the key is an object key,
     * the value is retrieved from the `_nodes` array using the index stored in the `OBJ_KEY_INDEX`
     * property of the key. If the key is a string key, the value is retrieved from the `_noObjMap` object
     * using the key itself. If the key is not found, `undefined` is
     */
    get(key: K): V | undefined;
    /**
     * Time Complexity: O(n), where n is the index.
     * Space Complexity: O(1)
     *
     * The function `getAt` retrieves the key-value pair at a specified index in a linked list.
     * @param {number} index - The index parameter is a number that represents the position of the
     * element we want to retrieve from the data structure.
     * @returns The method `getAt(index: number)` is returning an array containing the key-value pair at
     * the specified index in the data structure. The key-value pair is represented as a tuple `[K, V]`,
     * where `K` is the key and `V` is the value.
     */
    getAt(index: number): [K, V];
    /**
     * Time Complexity: O(1)
     * Space Complexity: O(1)
     *
     * The `delete` function removes a key-value pair from a map-like data structure.
     * @param {K} key - The `key` parameter is the key that you want to delete from the data structure.
     * It can be of any type, but typically it is a string or an object.
     * @returns a boolean value. It returns `true` if the deletion was successful, and `false` if the key
     * was not found.
     */
    delete(key: K): boolean;
    /**
     * Time Complexity: O(n), where n is the index.
     * Space Complexity: O(1)
     *
     * The `deleteAt` function deletes a node at a specified index in a linked list.
     * @param {number} index - The index parameter represents the position at which the node should be
     * deleted in the linked list.
     * @returns The size of the list after deleting the element at the specified index.
     */
    deleteAt(index: number): number;
    /**
     * Time Complexity: O(1)
     * Space Complexity: O(1)
     *
     * The function checks if a data structure is empty by comparing its size to zero.
     * @returns The method is returning a boolean value indicating whether the size of the object is 0 or
     * not.
     */
    isEmpty(): boolean;
    /**
     * Time Complexity: O(1)
     * Space Complexity: O(1)
     *
     * The `clear` function clears all the elements in a data structure and resets its properties.
     */
    clear(): void;
    /**
     * Time Complexity: O(n), where n is the number of elements in the HashMap.
     * Space Complexity: O(1)
     *
     * The `forEach` function iterates over each element in a HashMap and executes a callback function on
     * each element.
     * @param callback - The callback parameter is a function that will be called for each element in the
     * HashMap. It takes three arguments:
     */
    forEach(callback: (element: [K, V], index: number, hashMap: HashMap<K, V>) => void): void;
    /**
     * The `filter` function takes a predicate function and returns a new HashMap containing only the
     * key-value pairs that satisfy the predicate.
     * @param predicate - The `predicate` parameter is a function that takes two arguments: `element` and
     * `map`.
     * @returns a new HashMap object that contains the key-value pairs from the original HashMap that
     * satisfy the given predicate function.
     */
    filter(predicate: (element: [K, V], index: number, map: HashMap<K, V>) => boolean): HashMap<K, V>;
    /**
     * The `map` function takes a callback function and returns a new HashMap with the values transformed
     * by the callback.
     * @param callback - The `callback` parameter is a function that takes two arguments: `element` and
     * `map`.
     * @returns a new HashMap object with the values mapped according to the provided callback function.
     */
    map<NV>(callback: (element: [K, V], index: number, map: HashMap<K, V>) => NV): HashMap<K, NV>;
    /**
     * The `reduce` function iterates over the elements of a HashMap and applies a callback function to
     * each element, accumulating a single value.
     * @param callback - The callback parameter is a function that takes three arguments: accumulator,
     * element, and map. It is called for each element in the HashMap and is used to accumulate a single
     * result.
     * @param {A} initialValue - The `initialValue` parameter is the initial value of the accumulator. It
     * is the value that will be passed as the first argument to the `callback` function when reducing
     * the elements of the map.
     * @returns The `reduce` function is returning the final value of the accumulator after iterating
     * over all the elements in the HashMap and applying the callback function to each element.
     */
    reduce<A>(callback: (accumulator: A, element: [K, V], index: number, map: HashMap<K, V>) => A, initialValue: A): A;
    /**
     * Time Complexity: O(n), where n is the number of elements in the HashMap.
     * Space Complexity: O(1)
     *
     * The above function is an iterator that yields key-value pairs from a linked list.
     */
    [Symbol.iterator](): Generator<[K, V], void, unknown>;
    print(): void;
    /**
     * Time Complexity: O(1)
     * Space Complexity: O(1)
     *
     * The `_deleteNode` function removes a node from a doubly linked list and updates the head and tail
     * pointers if necessary.
     * @param node - The `node` parameter is an instance of the `HashMapLinkedNode` class, which
     * represents a node in a linked list. It contains a key-value pair and references to the previous
     * and next nodes in the list.
     */
    protected _deleteNode(node: HashMapLinkedNode<K, V | undefined>): void;
}
